package com.sf.service.gateway.service.impl;

import java.util.*;
import java.util.stream.Collectors;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.NumberUtil;
import com.alibaba.fastjson2.JSON;
import com.sf.common.utils.*;
import com.sf.service.gateway.domain.GatewayInterfaceInfo;
import com.sf.service.gateway.domain.GatewayInterfaceLinkStrategy;
import com.sf.service.gateway.domain.GatewayInterfaceMock;
import com.sf.service.gateway.domain.GatewayStrategy;
import com.sf.service.gateway.domain.dto.InsertGatewayInterfaceMockDTO;
import com.sf.vertx.api.pojo.MockMatchCondition;
import com.sf.service.gateway.domain.dto.InsertOrUpdateGatewayInterfaceInfoDTO;
import com.sf.service.gateway.enums.GatewayDataStatus;
import com.sf.service.gateway.enums.GatewayStrategyType;
import com.sf.service.gateway.mapper.GatewayInterfaceInfoMapper;
import com.sf.service.gateway.mapper.GatewayInterfaceLinkStrategyMapper;
import com.sf.service.gateway.mapper.GatewayInterfaceMockMapper;
import com.sf.service.gateway.service.IGatewayInterfaceInfoService;
import com.sf.service.gateway.service.IGatewayStrategyService;
import com.sf.service.index.utils.AppUtils;
import com.sf.vertx.enums.MatchType;
import lombok.extern.slf4j.Slf4j;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;

/**
 * 接口管理Service业务层处理
 *
 * @author zoukun
 * @date 2024-04-22
 */
@Slf4j
@Service
public class GatewayInterfaceInfoServiceImpl implements IGatewayInterfaceInfoService {

    @Resource
    private GatewayInterfaceInfoMapper gatewayInterfaceInfoMapper;

    @Resource
    private GatewayInterfaceLinkStrategyMapper gatewayInterfaceLinkStrategyMapper;

    @Resource
    private GatewayInterfaceMockMapper gatewayInterfaceMockMapper;

    @Resource
    private IGatewayStrategyService gatewayStrategyService;

    /**
     * 查询接口管理
     *
     * @param id 接口管理主键
     * @return 接口管理
     */
    @Override
    public GatewayInterfaceInfo selectGatewayInterfaceInfoById(Long id) {
        return gatewayInterfaceInfoMapper.selectGatewayInterfaceInfoById(id);
    }

    /**
     * 查询接口管理列表
     *
     * @param gatewayInterfaceInfo 接口管理
     * @return 接口管理
     */
    @Override
    public List<GatewayInterfaceInfo> selectGatewayInterfaceInfoList(GatewayInterfaceInfo gatewayInterfaceInfo) {
        return gatewayInterfaceInfoMapper.selectGatewayInterfaceInfoList(gatewayInterfaceInfo);
    }

    /**
     * 新增接口管理
     *
     * @param gatewayInterfaceInfo 接口管理
     * @return 结果
     */
    @Override
    public Long insertGatewayInterfaceInfo(GatewayInterfaceInfo gatewayInterfaceInfo) {
        checkGatewayInterfaceInfoCanInsertOrUpdate(gatewayInterfaceInfo);
        String appCode = AppUtils.getAppCodeFromRequestHeader();
        gatewayInterfaceInfo.setAppCode(appCode);
        String username = SecurityUtils.getUsername();
        gatewayInterfaceInfo.setCreated(username);
        gatewayInterfaceInfo.setModified(username);
        Date nowDate = DateUtils.getNowDate();
        gatewayInterfaceInfo.setCreateTime(nowDate);
        gatewayInterfaceInfo.setUpdateTime(nowDate);
        try {
            gatewayInterfaceInfoMapper.insertGatewayInterfaceInfo(gatewayInterfaceInfo);
            return gatewayInterfaceInfo.getId();
        } catch (DuplicateKeyException e) {
            throw new IllegalArgumentException("接口编码已存在!");
        }

    }

    private void checkGatewayInterfaceInfoCanInsertOrUpdate(GatewayInterfaceInfo gatewayInterfaceInfo) {
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayInterfaceInfo.getMockStatus())) {
            Assert.notNull(gatewayInterfaceInfo.getMockDefaultHttpStatus(), "请输入默认响应HTTP状态码");
            Assert.isTrue(JSONUtils.isJSONObject(gatewayInterfaceInfo.getMockDefaultResponse()), "请输入正确格式的默认响应信息");
        }
    }


    /**
     * 修改接口管理
     *
     * @param gatewayInterfaceInfo 接口管理
     * @return 结果
     */
    @Override
    public int updateGatewayInterfaceInfo(GatewayInterfaceInfo gatewayInterfaceInfo) {
        checkGatewayInterfaceInfoCanInsertOrUpdate(gatewayInterfaceInfo);
        gatewayInterfaceInfo.setModified(SecurityUtils.getUsername());
        gatewayInterfaceInfo.setUpdateTime(DateUtils.getNowDate());
        try {
            return gatewayInterfaceInfoMapper.updateGatewayInterfaceInfo(gatewayInterfaceInfo);
        } catch (DuplicateKeyException e) {
            throw new IllegalArgumentException("接口编码已存在!");
        }
    }

    /**
     * 批量删除接口管理
     *
     * @param ids 需要删除的接口管理主键
     * @return 结果
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public int deleteGatewayInterfaceInfoByIds(Long[] ids) {
        if (ids == null || ids.length == 0) {
            return 0;
        }
        gatewayInterfaceLinkStrategyMapper.deleteByInterfaceIds(CollUtil.newHashSet(ids));
        return gatewayInterfaceInfoMapper.deleteGatewayInterfaceInfoByIds(ids);
    }

    /**
     * 删除接口管理信息
     *
     * @param id 接口管理主键
     * @return 结果
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public int deleteGatewayInterfaceInfoById(Long id) {
        if (id == null) {
            return 0;
        }
        gatewayInterfaceLinkStrategyMapper.deleteByInterfaceIds(Collections.singleton(id));
        return gatewayInterfaceInfoMapper.deleteGatewayInterfaceInfoById(id);
    }

    @Override
    public int updateGatewayInterfaceInfoStatusByIds(Collection<Long> ids, GatewayDataStatus status) {
        if (CollectionUtils.isEmpty(ids) || status == null) {
            log.warn("Insufficient update conditions");
            return 0;
        }
        return gatewayInterfaceInfoMapper.updateGatewayInterfaceInfoStatusByIds(ids, status.getCode());
    }

    @Override
    public List<GatewayInterfaceInfo> selectGatewayInterfaceInfoByServerIds(Collection<Long> serverIds) {
        if (CollectionUtils.isEmpty(serverIds)) {
            return Collections.emptyList();
        }
        return gatewayInterfaceInfoMapper.selectGatewayInterfaceInfoByServerIds(serverIds);
    }

    @Override
    public List<GatewayInterfaceInfo> selectGatewayInterfaceInfoByIds(Collection<Long> ids) {
        if (CollectionUtils.isEmpty(ids)) {
            return Collections.emptyList();
        }
        return gatewayInterfaceInfoMapper.selectGatewayInterfaceInfoByIds(ids);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long insertGatewayInterfaceInfoAndBindStrategy(InsertOrUpdateGatewayInterfaceInfoDTO dto) {
        GatewayInterfaceInfo gatewayInterfaceInfo = dto.convertToGatewayInterfaceInfo();
        gatewayInterfaceInfo.setStatus(GatewayDataStatus.DISABLE.getCode());
        Long interfaceInfoId = insertGatewayInterfaceInfo(gatewayInterfaceInfo);
        if (CollUtil.isNotEmpty(dto.getStrategyIds())) {
            checkGatewayStrategyTypeCannotBeRepeated(dto.getStrategyIds());
            List<GatewayInterfaceLinkStrategy> linkStrategyList = dto.getStrategyIds()
                    .stream().map(strategyId -> new GatewayInterfaceLinkStrategy(interfaceInfoId, strategyId))
                    .collect(Collectors.toList());
            gatewayInterfaceLinkStrategyMapper.batchInsertGatewayInterfaceLinkStrategy(linkStrategyList);
        }
        if (CollUtil.isNotEmpty(dto.getMockExpectations())) {
            checkInsertGatewayInterfaceMockList(dto.getMockExpectations());
            List<GatewayInterfaceMock> mockList = dto.getMockExpectations().stream().map(mockDTO -> {
                GatewayInterfaceMock mock = new GatewayInterfaceMock();
                mock.setInterfaceId(interfaceInfoId);
                mock.setHttpStatus(mockDTO.getHttpStatus());
                mock.setMatchConditions(JSON.toJSONString(mockDTO.getMatchConditions()));
                mock.setMockResponse(mockDTO.getMockResponse());
                mock.setStatus(mockDTO.getStatus());
                mock.setOrderNum(mockDTO.getOrderNum());
                return mock;
            }).collect(Collectors.toList());
            gatewayInterfaceMockMapper.batchInsertGatewayInterfaceMock(mockList);
        }
        return interfaceInfoId;
    }

    private void checkInsertGatewayInterfaceMockList(List<InsertGatewayInterfaceMockDTO> gatewayInterfaceMockList) {
        gatewayInterfaceMockList.forEach(mock -> {
            Assert.notNull(mock.getHttpStatus(), "HTTP状态码不能为空");
            Assert.isTrue(JSONUtils.isJSONObject(mock.getMockResponse()), "请输入正确格式的默认响应信息");
            List<MockMatchCondition> conditionDTOList = mock.getMatchConditions();
            Assert.notEmpty(conditionDTOList, "至少存在一条匹配条件");
            conditionDTOList.forEach(condition -> {
                MatchType matchType = MatchType.getByCode(condition.getMatchType());
                Assert.notNull(matchType, "不支持的匹配类型");
                switch (matchType) {
                    case IS_NULL:
                    case NOT_NULL:
                        break;
                    case LT:
                    case LE:
                    case GT:
                    case GE:
                        Assert.notEmpty(condition.getParameterValue(), "请输入参数值");
                        Assert.isTrue(NumberUtil.isNumber(condition.getParameterValue().get(0)), "大于、大于或等于、小于、小于或等于类型，参数值只能是数字");
                        break;
                    default:
                        Assert.notEmpty(condition.getParameterValue(), "请输入参数值");
                        break;
                }
            });
        });
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long updateGatewayInterfaceInfoAndBindStrategy(InsertOrUpdateGatewayInterfaceInfoDTO dto) {
        GatewayInterfaceInfo gatewayInterfaceInfo = dto.convertToGatewayInterfaceInfo();
        updateGatewayInterfaceInfo(gatewayInterfaceInfo);
        // 先删除
        gatewayInterfaceLinkStrategyMapper.deleteByInterfaceIds(Collections.singleton(dto.getId()));
        if (CollUtil.isNotEmpty(dto.getStrategyIds())) {
            checkGatewayStrategyTypeCannotBeRepeated(dto.getStrategyIds());
            List<GatewayInterfaceLinkStrategy> linkStrategyList = dto.getStrategyIds()
                    .stream().map(strategyId -> new GatewayInterfaceLinkStrategy(dto.getId(), strategyId))
                    .collect(Collectors.toList());
            gatewayInterfaceLinkStrategyMapper.batchInsertGatewayInterfaceLinkStrategy(linkStrategyList);
        }
        gatewayInterfaceMockMapper.deleteByInterfaceIds(Collections.singleton(dto.getId()));
        if (CollUtil.isNotEmpty(dto.getMockExpectations())) {
            checkInsertGatewayInterfaceMockList(dto.getMockExpectations());
            List<GatewayInterfaceMock> mockList = dto.getMockExpectations().stream().map(mockDTO -> {
                GatewayInterfaceMock mock = new GatewayInterfaceMock();
                mock.setInterfaceId(dto.getId());
                mock.setHttpStatus(mockDTO.getHttpStatus());
                mock.setMatchConditions(JSON.toJSONString(mockDTO.getMatchConditions()));
                mock.setMockResponse(mockDTO.getMockResponse());
                mock.setStatus(mockDTO.getStatus());
                mock.setOrderNum(mockDTO.getOrderNum());
                return mock;
            }).collect(Collectors.toList());
            gatewayInterfaceMockMapper.batchInsertGatewayInterfaceMock(mockList);
        }
        return gatewayInterfaceInfo.getId();
    }

    @Override
    public List<GatewayInterfaceLinkStrategy> selectGatewayInterfaceLinkStrategyByInterfaceIds(Collection<Long> interfaceIds) {
        if (CollectionUtils.isEmpty(interfaceIds)) {
            return Collections.emptyList();
        }
        return gatewayInterfaceLinkStrategyMapper.selectGatewayInterfaceLinkStrategyByInterfaceIds(interfaceIds);
    }

    @Override
    public List<GatewayInterfaceMock> selectGatewayInterfaceMockListByInterfaceIds(Collection<Long> interfaceIds) {
        if (CollectionUtils.isEmpty(interfaceIds)) {
            return Collections.emptyList();
        }
        return gatewayInterfaceMockMapper.selectGatewayInterfaceMockByInterfaceIds(interfaceIds);

    }

    private void checkGatewayStrategyTypeCannotBeRepeated(List<Long> strategyIds) {
        List<GatewayStrategy> strategyList = gatewayStrategyService.selectGatewayStrategyByIds(strategyIds);
        Map<String, List<GatewayStrategy>> groupByStrategyTypeMap = strategyList.stream().collect(Collectors.groupingBy(GatewayStrategy::getStrategyType));
        groupByStrategyTypeMap.forEach((type, strategies) -> {
            Assert.isTrue(strategies.size() == 1, Objects.requireNonNull(GatewayStrategyType.getByCode(type)).getInfo() + "类型只能选择一个！");
        });
    }

}
